/* 
 * File:   gdgelement.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 13. května 2019, 22:40
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdint.h>
#include <glib.h>
#include <glib/gstdio.h>

#include "gdgelement.h"
#include "gdgpart.h"


static char* gdgelement_create_txt_field ( const char *txt ) {
    int len = strlen ( txt );
    char *txt_field = g_new0 ( char, len + 1 );
    strncpy ( txt_field, txt, len );
    return txt_field;
}


static void gdgelement_add_port ( st_GDGELEMENT *elm, const char *name, const char *search_name, en_GDGELEMENT_PORT_TYPE type ) {

    st_GDGELEMENT_PORT *port = g_new0 ( st_GDGELEMENT_PORT, 1 );
    port->name = gdgelement_create_txt_field ( name );
    port->search_name = gdgelement_create_txt_field ( search_name );
    port->type = type;

    int i = elm->count_port;
    elm->count_port++;
    if ( !i ) {
        elm->port = g_new0 ( st_GDGELEMENT_PORT*, 1 );
    } else {
        elm->port = g_renew ( st_GDGELEMENT_PORT*, elm->port, elm->count_port );
    };
    //printf("%s_%s : add new port: '%s' : %s\n", elm->name, elm->description, port->name, (port->type == GDGENTITY_PORT_IN) ? "In" : "Out");
    port->id = i;
    elm->port[i] = port;
}


static inline void gdgelement_add_input_port ( st_GDGELEMENT *elm, const char *name, const char *search_name ) {
    gdgelement_add_port ( elm, name, search_name, GDGELEMENT_PORT_IN );
}


static inline void gdgelement_add_output_port ( st_GDGELEMENT *elm, const char *name, const char *search_name ) {
    gdgelement_add_port ( elm, name, search_name, GDGELEMENT_PORT_OUT );
}


static st_GDGELEMENT* gdgelement_new ( const char *name, const char *description ) {
    st_GDGELEMENT *elm = g_new0 ( st_GDGELEMENT, 1 );
    elm->name = gdgelement_create_txt_field ( name );
    elm->description = gdgelement_create_txt_field ( description );
    //printf("Created element %s_%s\n", elm->name, elm->description);
    return elm;
}


static void gdgelement_port_destroy ( st_GDGELEMENT_PORT *port ) {
    if ( !port ) return;
    g_free ( port->name );
    g_free ( port->search_name );
    g_free ( port );
}


static void gdgelement_destroy ( st_GDGELEMENT *elm ) {
    if ( !elm ) return;
    int i;
    for ( i = 0; i < elm->count_port; i++ ) {
        gdgelement_port_destroy ( elm->port[i] );
    };

    g_free ( elm->name );
    g_free ( elm->description );
    g_free ( elm );
}


void gdgelements_destroy ( st_GDGELEMENTS_ALL *elmall ) {
    if ( !elmall ) return;
    int i;
    for ( i = 0; i < elmall->count_element; i++ ) {
        gdgelement_destroy ( elmall->element[i] );
    };
    g_free ( elmall );
}


st_GDGELEMENT* gdgelements_search_element_by_name ( st_GDGELEMENTS_ALL *ent, const char *name ) {
    int i;
    for ( i = 0; i < ent->count_element; i++ ) {
        st_GDGELEMENT *elm = ent->element[i];
        if ( 0 == strcmp ( elm->name, name ) ) {
            return elm;
        };
    };
    fprintf ( stderr, "%s(%d) - ERROR: element '%s' not found!\n", __func__, __LINE__, name );
    return NULL;
}


static void gdgelements_add_element ( st_GDGELEMENTS_ALL *ent, st_GDGELEMENT *elm ) {
    int i = ent->count_element;
    ent->count_element++;
    if ( !i ) {
        ent->element = g_new0 ( st_GDGELEMENT*, 1 );
    } else {
        ent->element = g_renew ( st_GDGELEMENT*, ent->element, ent->count_element );
    };
    ent->element[i] = elm;
}


void gdgelement_register_part_instance ( st_GDGELEMENT *elm, void *part ) {
    elm->count_instance++;
}


st_GDGELEMENT_PORT* gdgelement_get_port_by_name ( st_GDGELEMENT *elm, char *port_name ) {

    int i;
    for ( i = 0; i < elm->count_port; i++ ) {
        st_GDGELEMENT_PORT *port = elm->port[i];
        if ( 0 == strcmp ( port->search_name, port_name ) ) return port;
    };

    return NULL;
}


st_GDGELEMENTS_ALL* gdgelements_init ( void ) {
    st_GDGELEMENTS_ALL *elmall = g_new0 ( st_GDGELEMENTS_ALL, 1 );

    st_GDGELEMENT *elm;

    elm = gdgelement_new ( "F100", "NOT" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F101", "NOT_NOT" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F102", "NOT16" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F103", "NOT24" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F104", "NOT32" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F105", "NOT16_NOT16" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F111", "Buffer" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F112", "Buffer16" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F113", "Buffer24" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F114", "Buffer32" );
    gdgelement_add_input_port ( elm, "I", "IN" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F202", "NOR" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F203", "NOR3" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F204", "NOR4" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F212", "OR" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F301", "NAND_NAND" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F302", "NAND" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F303", "NAND3" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F304", "NAND4" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F306", "NAND6" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_input_port ( elm, "I5", "IN5" );
    gdgelement_add_input_port ( elm, "I6", "IN6" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F312", "AND" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F421", "AND_NOR_1_2" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F422", "AND_NOR_1_1_2" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F423", "AND_NOR_1_3" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F424", "AND_NOR_2_2" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F431", "OR_NAND_1_2" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F433", "OR_NAND_1_3" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F434", "OR_NAND_2_2" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F512", "NXOR" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F521", "1_Bit_Full_Adder" );
    gdgelement_add_input_port ( elm, "A", "INA" );
    gdgelement_add_input_port ( elm, "B", "INB" );
    gdgelement_add_input_port ( elm, "CI", "CI" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "CO", "CO" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F601", "D_Latch" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F604", "D_Latch_C" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "nCLK", "CLK" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F611", "D_Type_Flip_Flop" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F612", "D_Type_Flip_Flop_Reset" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "RES", "RST" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F615", "D_Type_Flip_Flop_Inv_Reset" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "nRES", "RST" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F616", "D_Type_Flip_Flop_Inv_Set" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "nSET", "SET" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F617", "D_Type_Flip_Flop_Inv_Set_Inv_Reset" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "nSET", "SET" );
    gdgelement_add_input_port ( elm, "nRES", "RST" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F631", "D_Type_Flip_Flop_C" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "nCLK", "CLK" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F635", "D_Type_Flip_Flop_C_Inv_Reset" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "nCLK", "CLK" );
    gdgelement_add_input_port ( elm, "nRES", "RST" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F641", "D_Type_Buffered_Flip_Flop" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F642", "D_Type_Buffered_Flip_Flop_Reset" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "RES", "RST" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F645", "D_Type_Buffered_Flip_Flop_Inv_Reset" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "nRES", "RST" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F666", "D_Type_Buffered_Flip_Flop_Inv_Set" );
    gdgelement_add_input_port ( elm, "I", "DATA" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "nSET", "SET" );
    gdgelement_add_output_port ( elm, "O", "OUT" );
    gdgelement_add_output_port ( elm, "nO", "OUT2" );
    gdgelements_add_element ( elmall, elm );

    elm = gdgelement_new ( "F961", "4_Bit_Counter" );
    gdgelement_add_input_port ( elm, "CLK", "CLK" );
    gdgelement_add_input_port ( elm, "I1", "IN1" );
    gdgelement_add_input_port ( elm, "I2", "IN2" );
    gdgelement_add_input_port ( elm, "I3", "IN3" );
    gdgelement_add_input_port ( elm, "I4", "IN4" );
    gdgelement_add_input_port ( elm, "CARRY_IN", "CI" );
    gdgelement_add_input_port ( elm, "ENABLE", "EN" );
    gdgelement_add_input_port ( elm, "nSET", "SET" );
    gdgelement_add_input_port ( elm, "nRES", "RST" );
    gdgelement_add_output_port ( elm, "O1", "OUT1" );
    gdgelement_add_output_port ( elm, "O2", "OUT2" );
    gdgelement_add_output_port ( elm, "O3", "OUT3" );
    gdgelement_add_output_port ( elm, "O4", "OUT4" );
    gdgelement_add_output_port ( elm, "nCO1", "CO1" );
    gdgelement_add_output_port ( elm, "nCO2", "CO2" );
    gdgelement_add_output_port ( elm, "nCO3", "CO3" );
    gdgelement_add_output_port ( elm, "nCO4", "CO4" );
    gdgelement_add_output_port ( elm, "nCARRY_OUT", "CO" );
    gdgelements_add_element ( elmall, elm );

    fprintf ( stderr, "%s() - initialised %d elements\n", __func__, elmall->count_element );

    return elmall;
}
