/* 
 * File:   gdgnet.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 13. května 2019, 21:44
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdint.h>
#include <glib.h>

#include "gdgnet.h"
#include "main.h"


static char* gdgnet_create_txt_field ( const char *txt ) {
    int len = strlen ( txt );
    char *txt_field = g_new0 ( char, len + 1 );
    strncpy ( txt_field, txt, len );
    return txt_field;
}


static void gdgnet_pin_destroy ( st_GDGNET_PIN *pin ) {
    if ( !pin ) return;
    if ( pin->part_pin_name ) g_free ( pin->part_pin_name );
    g_free ( pin );
}


static st_GDGNET_PIN* gdgnet_pin_create ( st_GDGELEMENT *elm, const char *part_pin_name, int loc_x, int loc_y, int loc_z ) {
    st_GDGNET_PIN *pin = g_new0 ( st_GDGNET_PIN, 1 );
    pin->element = elm;
    pin->part_pin_name = gdgnet_create_txt_field ( part_pin_name );
    pin->loc_x = loc_x;
    pin->loc_y = loc_y;
    pin->loc_z = loc_z;
    //printf ( "net connect: %s, %s\n", ( elm ) ? elm->name : "PIN", pin->part_pin_name );
    return pin;
}


static void gdgnet_add_pin ( st_GDGNET *net, st_GDGNET_PIN *pin ) {
    int i = net->count_pin;
    net->count_pin++;
    if ( !i ) {
        net->pin = g_new0 ( st_GDGNET_PIN*, 1 );
    } else {
        net->pin = g_renew ( st_GDGNET_PIN*, net->pin, net->count_pin );
    };
    net->pin[i] = pin;
}


static st_GDGNET* gdgnet_create ( const char *name ) {
    st_GDGNET *net = g_new0 ( st_GDGNET, 1 );

    if ( 0 == strcmp ( name, "0" ) ) {
        name = "const_GND";
        net->is_constant = TRUE;
        net->c_value = FALSE;
    } else if ( 0 == strcmp ( name, "vdd" ) ) {
        name = "const_VDD";
        net->is_constant = TRUE;
        net->c_value = TRUE;
    } else {
        net->is_constant = FALSE;
        net->c_value = FALSE;
    };
    
    net->name = gdgnet_create_txt_field ( name );
    //printf ( "create NET: %s\n", name );
    return net;
}


static void gdgnet_destroy ( st_GDGNET *net ) {
    if ( !net ) return;
    if ( net->name ) g_free ( net->name );
    int i;
    for ( i = 0; i < net->count_pin; i++ ) {
        gdgnet_pin_destroy ( net->pin[i] );
    };
    if ( net->pin ) g_free ( net->pin );
    g_free ( net );
}


void gdgnets_destroy ( st_GDGNETS_ALL *netall ) {
    if ( !netall ) return;
    int i;
    for ( i = 0; i < netall->count_net; i++ ) {
        gdgnet_destroy ( netall->net[i] );
    };
    if ( netall->net ) g_free ( netall->net );
    g_free ( netall );
}


static void gdgnets_add_net ( st_GDGNETS_ALL *netall, st_GDGNET *net ) {
    int i = netall->count_net;
    netall->count_net++;
    if ( !i ) {
        netall->net = g_new0 ( st_GDGNET*, 1 );
    } else {
        netall->net = g_renew ( st_GDGNET*, netall->net, netall->count_net );
    };
    netall->net[i] = net;
}


static gboolean gdgnet_check ( st_GDGNETS_ALL *netall ) {

    gboolean result = TRUE;

    int i;
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];

        if ( net->count_output > 1 ) {
            fprintf ( stderr, "Error - NET '%s' has %d outputs\n", net->name, net->count_output );
            result = FALSE;
        };

        /*
                if ( ( net->count_output == 0 ) && ( net->count_externals == 0 ) ) {
                    fprintf ( stderr, "Error - NET '%s' have not output pins\n", net->name, net->count_output );
                    result = FALSE;
                };
         */

        /*
                if ( net->count_externals > 1 ) {
                    fprintf ( stderr, "Error - NET '%s' have '%d' externals\n", net->name, net->count_externals );
                    result = FALSE;
                };
         */

    };

    return result;
}


/**
 * 
 * Parts:
 * 
 *  Fnnn_X_Y
 *  PIN_X_Y
 *  PIN_X_Y_Z
 * 
 */
st_GDGNETS_ALL* gdgnets_init ( st_GDGELEMENTS_ALL *elmall, const char * filepath ) {

    FILE *fp = fopen ( filepath, "r" );
    if ( !fp ) {
        GString *str = g_string_new ( "" );
        g_string_append_printf ( str, "%s(%d) - ERROR: Can't open input file '%s'\n", __func__, __LINE__, filepath );
        main_error_exit ( g_string_free ( str, FALSE ) );
        return NULL;
    };

    st_GDGNETS_ALL *netall = g_new0 ( st_GDGNETS_ALL, 1 );

    char buf[1024];
    int row = 0;

    // prvni radky preskocime
    GRegex *regex_netlist = g_regex_new ( "^Net\\s+Part\\s+Pin\\s+Net", 0, 0, NULL );
    while ( fgets ( buf, sizeof ( buf ), fp ) ) {
        row++;
        GMatchInfo *match_info;
        if ( g_regex_match ( regex_netlist, buf, 0, &match_info ) ) {
            g_match_info_free ( match_info );
            break;
        };
        g_match_info_free ( match_info );
    };
    g_regex_unref ( regex_netlist );

    //printf ( "%s() - NET search starting on %d row\n", __func__, row );

    GRegex *regex = g_regex_new ( "^([A-Za-z0-9_]+|\\s+)\\s+(\\w+)\\s+(\\w+)\\s+([GDS])", 0, 0, NULL );
    GRegex *regex_pname = g_regex_new ( "^(F\\d+|PIN)_(\\d+)_(\\d+)(_\\d+){0,1}", 0, 0, NULL );

    st_GDGNET *net = NULL;

    while ( fgets ( buf, sizeof ( buf ), fp ) ) {

        GMatchInfo *match_info;

        if ( g_regex_match ( regex, buf, 0, &match_info ) ) {
            //printf ( "%s", buf );

            gchar *net_name = g_match_info_fetch ( match_info, 1 );
            gchar *part_name = g_match_info_fetch ( match_info, 2 );
            gchar *pin_name = g_match_info_fetch ( match_info, 3 );

            if ( net_name[0] != ' ' ) {
                // novy net
                //printf ( "%s", buf );
                if ( net != NULL ) {
                    fprintf ( stderr, "%s(%d) - ERROR: net is not NULL! (row: %d)\n", __func__, __LINE__, row );
                    return NULL;
                };

                net = gdgnet_create ( net_name );
                gdgnets_add_net ( netall, net );

            } else {
                // novy part pin
                //printf ( "%s", buf );
                if ( net == NULL ) {
                    fprintf ( stderr, "%s(%d) - ERROR: net is NULL! (row: %d)\n", __func__, __LINE__, row );
                    return NULL;
                };
            };

            // TODO: analyzovat part name

            GMatchInfo *match_info_pname;
            if ( !g_regex_match ( regex_pname, part_name, 0, &match_info_pname ) ) {
                fprintf ( stderr, "%s", buf );
                fprintf ( stderr, "%s(%d) - ERROR: bad part name! (row: %d)\n", __func__, __LINE__, row );
                return NULL;
            };

            gchar *part_element_name = g_match_info_fetch ( match_info_pname, 1 );
            gchar *loc_x_txt = g_match_info_fetch ( match_info_pname, 2 );
            gchar *loc_y_txt = g_match_info_fetch ( match_info_pname, 3 );
            gchar *loc_z_txt = g_match_info_fetch ( match_info_pname, 4 );

            st_GDGELEMENT *elm = NULL; // zustane NULL, pokud se jedna o PIN
            if ( 0 != strcmp ( part_element_name, "PIN" ) ) {
                elm = gdgelements_search_element_by_name ( elmall, part_element_name );
                if ( !elm ) {
                    fprintf ( stderr, "%s", buf );
                    fprintf ( stderr, "%s(%d) - ERROR: element not found! (row: %d)\n", __func__, __LINE__, row );
                    return NULL;
                };
            };

            int loc_x = atoi ( loc_x_txt );
            int loc_y = atoi ( loc_y_txt );
            int loc_z = ( loc_z_txt ) ? atoi ( &loc_z_txt[1] ) : 0;

            // pridat pin
            st_GDGNET_PIN *pin = gdgnet_pin_create ( elm, pin_name, loc_x, loc_y, loc_z );
            gdgnet_add_pin ( net, pin );

            if ( elm != NULL ) {
                if ( ( 0 == strncmp ( pin_name, "OUT", 3 ) ) || ( 0 == strncmp ( pin_name, "CO", 2 ) ) ) {
                    net->count_output++;
                };
            } else {
                net->count_externals++;
            };

            /*
                        if ( ( !elm ) && ( loc_z != 0 ) ) {
                            printf ( "NET %d: %s\n", loc_z, net->name );
                        };
             */

            /*
                        if ( ( !elm ) && ( loc_z != 0 ) ) {
                            printf ( "NET %s - %s\n", net->name, pin_name );
                        };
             */

            g_match_info_free ( match_info_pname );

            g_free ( part_element_name );
            g_free ( loc_x_txt );
            g_free ( loc_y_txt );
            g_free ( loc_z_txt );

            g_free ( net_name );
            g_free ( part_name );
            g_free ( pin_name );

        } else {
            // Pokud jsme pracovali s nejakym net, tak uz jeho definice skoncila
            net = NULL;
        };

        g_match_info_free ( match_info );
        row++;
    };

    g_regex_unref ( regex );

    fclose ( fp );

    //printf ( "%s() - parsed %d input rows\n", __func__, row );
    fprintf ( stderr, "%s() - initialised %d nets\n", __func__, netall->count_net );

    if ( !gdgnet_check ( netall ) ) return NULL;

    return netall;
}


static st_GDGIOPORT* gdgnets_search_ioport ( st_GDGIOPORTS_ALL *ioportall, int loc_x, int loc_y, int loc_z ) {

    st_GDGIOPORT **ioportbank = NULL;
    int count = 0;

    if ( loc_z ) {
        ioportbank = ioportall->real;
        count = ioportall->count_real;
    } else {
        ioportbank = ioportall->internal;
        count = ioportall->count_internal;
    };

    int i;
    for ( i = 0; i < count; i++ ) {
        st_GDGNET_PIN *pin = ioportbank[i]->pin;
        if ( ( pin->loc_x == loc_x ) && ( pin->loc_y == loc_y ) && ( pin->loc_z == loc_z ) ) return ioportbank[i];
    };

    return NULL;
}


static void gdgnets_ioport_add ( st_GDGIOPORTS_ALL *ioportall, st_GDGNET *net, st_GDGNET_PIN *pin ) {

    st_GDGIOPORT *ioport = g_new0 ( st_GDGIOPORT, 1 );
    ioport->net = net;
    ioport->pin = pin;

    GString *name_str = g_string_new ( "" );

    if ( pin->loc_z ) {
        int i = ioportall->count_real;
        ioportall->count_real++;
        ioportall->real = g_renew ( st_GDGIOPORT*, ioportall->real, ioportall->count_real );
        ioportall->real[i] = ioport;

        g_string_append_printf ( name_str, "%s_%s", ( net->count_output ) ? "o" : "i", net->name );
        ioport->vhdl_name = g_string_free ( name_str, FALSE );

    } else {
        int i = ioportall->count_internal;
        ioportall->count_internal++;
        ioportall->internal = g_renew ( st_GDGIOPORT*, ioportall->internal, ioportall->count_internal );
        ioportall->internal[i] = ioport;

        g_string_append_printf ( name_str, "%s_PIN_%d_%d_%s", ( net->count_output ) ? "o" : "i", pin->loc_x, pin->loc_y, net->name );
        ioport->vhdl_name = g_string_free ( name_str, FALSE );
    };
}


void gdgnets_ioports_destroy ( st_GDGIOPORTS_ALL *ioportall ) {
    if ( !ioportall ) return;
    int i;
    for ( i = 0; i < ioportall->count_real; i++ ) {
        if ( ioportall->real[i]->vhdl_name ) g_free ( ioportall->real[i]->vhdl_name );
        g_free ( ioportall->real[i] );
    };
    g_free ( ioportall->real );
    for ( i = 0; i < ioportall->count_internal; i++ ) {
        if ( ioportall->internal[i]->vhdl_name ) g_free ( ioportall->internal[i]->vhdl_name );
        g_free ( ioportall->internal[i] );
    };
    g_free ( ioportall->internal );
    g_free ( ioportall );
}


static int gdgnets_ioport_cmpfunc ( const void *a, const void *b ) {
    st_GDGIOPORT **aa = (st_GDGIOPORT**) a;
    st_GDGIOPORT **bb = (st_GDGIOPORT**) b;
    st_GDGIOPORT *ioport_a = aa[0];
    st_GDGIOPORT *ioport_b = bb[0];
    return ( ioport_a->pin->loc_z - ioport_b->pin->loc_z );
}


st_GDGIOPORTS_ALL* gdgnets_create_ioports ( st_GDGNETS_ALL *netall ) {

    st_GDGIOPORTS_ALL *ioportall = g_new0 ( st_GDGIOPORTS_ALL, 1 );
    ioportall->real = g_new0 ( st_GDGIOPORT*, 1 );
    ioportall->internal = g_new0 ( st_GDGIOPORT*, 1 );

    int i;
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];
        int j;
        for ( j = 0; j < net->count_pin; j++ ) {
            st_GDGNET_PIN *pin = net->pin[j];

            if ( !pin->element ) {
                if ( !gdgnets_search_ioport ( ioportall, pin->loc_x, pin->loc_y, pin->loc_z ) ) gdgnets_ioport_add ( ioportall, net, pin );
            };
        };
    };

    qsort ( ioportall->real, ioportall->count_real, sizeof ( st_GDGIOPORT* ), gdgnets_ioport_cmpfunc );

    return ioportall;
}
