/* 
 * File:   main.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 12. května 2019, 21:55
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <glib.h>

#include "gdgelement.h"
#include "gdgpart.h"
#include "gdgnet.h"
#include "vhdl.h"
#include "main.h"


static gboolean main_connect_parts ( st_GDGPARTS_ALL *partall, st_GDGNETS_ALL *netall ) {

    int i;
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];

        int j;
        for ( j = 0; j < net->count_pin; j++ ) {
            st_GDGNET_PIN *pin = net->pin[j];

            // pripojujeme jen parts (instance elementu), nikoliv PIN
            if ( pin->element ) {

                st_GDGPART *part = gdgparts_get_part_by_location ( partall, pin->loc_x, pin->loc_y );

                if ( !part ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %d, pin: %d) can't found any part on this location (%s_%d_%d) for net: %s\n", __func__, __LINE__, i, j, pin->element->name, pin->loc_x, pin->loc_y, net->name );
                    return FALSE;
                    //continue;
                };

                if ( part->element != pin->element ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %d, pin: %d) found diferrent element (%s) on location %s_%d_%d for net: %s\n", __func__, __LINE__, i, j, part->element->name, pin->element->name, pin->loc_x, pin->loc_y, net->name );
                    return FALSE;
                    //continue;
                };
                //printf ( "OK net: %d, pin: %d found element (%d) on location %s_%d_%d for net: %s\n", i, j, part->element->name, pin->element->name, pin->loc_x, pin->loc_y, net->name );

                st_GDGELEMENT_PORT *port = gdgelement_get_port_by_name ( pin->element, pin->part_pin_name );
                if ( !port ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %s, pin: %d, %s_%d_%d pin %s not found\n", __func__, __LINE__, net->name, j, pin->element->name, pin->loc_x, pin->loc_y, pin->part_pin_name );
                    return FALSE;
                    //continue;
                };

                if ( part->netpin[port->id] ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %s, pin: %d, %s_%d_%d pin %s is already connected to net: %s\n", __func__, __LINE__, net->name, j, pin->element->name, pin->loc_x, pin->loc_y, pin->part_pin_name, part->netpin[port->id]->name );
                    return FALSE;
                    //continue;
                };

                // pripojime pin soucastky
                part->netpin[port->id] = net;
            };
        };

    };

    return TRUE;
}


static gboolean main_error_on_part_connections ( st_GDGPARTS_ALL *partall ) {
    int i;
    for ( i = 0; i < partall->count_part; i++ ) {
        st_GDGPART *part = partall->part[i];
        int j;
        for ( j = 0; j < part->element->count_port; j++ ) {
            if ( !part->netpin[j] ) {
                fprintf ( stderr, "%s(%d) - ERROR not connected port %s (%d) on part: %s_%d_%d\n", __func__, __LINE__, part->element->port[j]->search_name, j, part->element->name, part->loc_x, part->loc_y );
                return TRUE;
            };
        };
    };
    return FALSE;
}

static char *spice_filepath = "testGDG_mod.spice";
static char *netlist_filepath = "testGDG_mod_exp.txt";
static char *vhdl_filepath = NULL;

static GOptionEntry entries[] = {
    { "spice", 's', 0, G_OPTION_ARG_FILENAME, &spice_filepath, "SPICE file (testGDG_mod.spice)", "filepath" },
    { "nets", 'n', 0, G_OPTION_ARG_FILENAME, &netlist_filepath, "Netlist file (testGDG_mod_exp.txt)", "filepath" },
    { "output", 'o', 0, G_OPTION_ARG_FILENAME, &vhdl_filepath, "VHDL output file - using stdout if is not set", "filepath.vhd" },
    { NULL }
};

GOptionContext *context;


void main_error_exit ( char *error_message ) {
    fprintf ( stderr, "\n%s\n\n", error_message );
    fprintf ( stderr, g_option_context_get_help ( context, TRUE, NULL ) );
    exit ( EXIT_FAILURE );
}


static void main_print_version ( void ) {
    GString *str = g_string_new ( "" );
    g_string_append_printf ( str, "%s, %s\n", MAIN_PROGRAM_NAME, MAIN_PROGRAM_VERSION );
    fprintf ( stderr, str->str );
    g_string_free ( str, TRUE );
}


/*
 * 
 */
int main ( int argc, char** argv ) {

    main_print_version ( );

    GError *error = NULL;
    context = g_option_context_new ( "- convertor of GDG spice files to VHDL" );
    g_option_context_add_main_entries ( context, entries, NULL );
    if ( !g_option_context_parse ( context, &argc, &argv, &error ) ) {
        g_print ( "option parsing failed: %s\n", error->message );
        exit ( 1 );
    };

    st_GDGELEMENTS_ALL *elmall = gdgelements_init ( );
    if ( !elmall ) return EXIT_FAILURE;

    st_GDGPARTS_ALL *partall = gdgparts_init ( elmall, spice_filepath );
    if ( !partall ) return EXIT_FAILURE;

    st_GDGNETS_ALL *netall = gdgnets_init ( elmall, netlist_filepath );
    if ( !netall ) return EXIT_FAILURE;

    st_GDGIOPORTS_ALL *ioportall = gdgnets_create_ioports ( netall );

    gboolean err = main_connect_parts ( partall, netall ) ? FALSE : TRUE;
    err |= main_error_on_part_connections ( partall );

    if ( !err ) {

        const char *vhdl_name = "GDG_internal";
        const char *vhdl_arch = "RTL";
        const char *net_prefix = "net_";
        const char *part_prefix = "x";

        GString *str = g_string_new ( "" );
        char *txt;

        txt = vhdl_create_entity_header ( ioportall, vhdl_name );
        g_string_append ( str, txt );
        g_free ( txt );

        g_string_append_printf ( str, "\n\narchitecture %s of %s is\n\n", vhdl_arch, vhdl_name );

        txt = vhdl_create_components ( elmall, FALSE );
        g_string_append ( str, txt );
        g_free ( txt );

        txt = vhdl_create_signal_definitions ( netall, net_prefix );
        g_string_append ( str, txt );
        g_free ( txt );

        g_string_append_printf ( str, "\nbegin\n\n", vhdl_arch );

        txt = vhdl_create_entity_port_connections ( ioportall, netall, net_prefix );
        g_string_append ( str, txt );
        g_free ( txt );

        txt = vhdl_create_component_instances ( partall, part_prefix, net_prefix );
        g_string_append ( str, txt );
        g_free ( txt );

        g_string_append_printf ( str, "\nend %s;\n\n", vhdl_arch );

        if ( !vhdl_filepath ) {
            printf ( "%s", str->str );
        } else {
            FILE *fp = fopen ( vhdl_filepath, "w" );
            if ( !fp ) {
                GString *str = g_string_new ( "" );
                g_string_append_printf ( str, "%s(%d) - ERROR: Can't open output file '%s'\n", __func__, __LINE__, vhdl_filepath );
                main_error_exit ( g_string_free ( str, FALSE ) );
                return EXIT_FAILURE;
            };

            fprintf ( fp, "%s", str->str );
            fclose ( fp );
        };

        g_string_free ( str, TRUE );

    };

    gdgparts_destroy ( partall );
    gdgnets_ioports_destroy ( ioportall );
    gdgnets_destroy ( netall );
    gdgelements_destroy ( elmall );

    return EXIT_SUCCESS;
}

