/* 
 * File:   vhdl.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 15. května 2019, 22:05
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <glib.h>

#include "gdgelement.h"
#include "gdgpart.h"
#include "gdgnet.h"
#include "vhdl.h"
#include "main.h"


char* vhdl_create_entity_header ( st_GDGIOPORTS_ALL *ioportall, const char * name ) {

    GString *str = g_string_new ( "" );

    g_string_append_printf ( str, "\n--\n-- Created by: %s, %s\n", MAIN_PROGRAM_NAME, MAIN_PROGRAM_VERSION );

    GTimeZone *tz = g_time_zone_new_local ( );
    GDateTime *datetime = g_date_time_new_now ( tz );

    g_string_append_printf ( str, "-- Pased: %d-%02d-%02d %02d:%02d:%02d\n",
                             g_date_time_get_year ( datetime ),
                             g_date_time_get_month ( datetime ),
                             g_date_time_get_day_of_month ( datetime ),
                             g_date_time_get_hour ( datetime ),
                             g_date_time_get_minute ( datetime ),
                             g_date_time_get_second ( datetime ) );

    g_string_append ( str, "--\n\n" );

    g_time_zone_unref ( tz );
    g_date_time_unref ( datetime );
    
    g_string_append ( str, "library IEEE;\n" );
    g_string_append ( str, "use IEEE.STD_LOGIC_1164.ALL;\n\n" );

    g_string_append_printf ( str, "entity %s is\n", name );
    g_string_append ( str, "\tPort (\n" );

    g_string_append ( str, "\n" );
    g_string_append ( str, "\t\t-- Real pins\n" );
    g_string_append ( str, "\n" );

    int i;
    for ( i = 0; i < ioportall->count_real; i++ ) {
        st_GDGIOPORT *ioport = ioportall->real[i];
        st_GDGNET_PIN *pin = ioport->pin;
        st_GDGNET *net = ioport->net;

        g_string_append_printf ( str, "\t\t%s : %s STD_LOGIC%s\t-- pin: %d\n",
                                 ioport->vhdl_name,
                                 ( net->count_output ) ? "out" : "in",
                                 ( ( !ioportall->count_internal ) && ( i == ioportall->count_real - 1 ) ) ? "" : ";",
                                 pin->loc_z );
    };

    g_string_append ( str, "\n" );
    g_string_append ( str, "\t\t-- Internal pins\n" );
    g_string_append ( str, "\n" );


    for ( i = 0; i < ioportall->count_internal; i++ ) {
        st_GDGIOPORT *ioport = ioportall->internal[i];
        st_GDGNET *net = ioport->net;

        g_string_append_printf ( str, "\t\t%s : %s STD_LOGIC%s\n",
                                 ioport->vhdl_name,
                                 ( net->count_output ) ? "out" : "in",
                                 ( i == ioportall->count_internal - 1 ) ? "" : ";" );
    };

    g_string_append ( str, "\t);\n\n" );
    g_string_append_printf ( str, "end %s;\n", name );

    return g_string_free ( str, FALSE );
}


static const char* vhdl_get_port_type_txt ( en_GDGELEMENT_PORT_TYPE type ) {
    switch ( type ) {
        case GDGELEMENT_PORT_IN:
            return "IN";

        case GDGELEMENT_PORT_OUT:
            return "OUT";

        case GDGELEMENT_PORT_INOUT:
            return "INOUT";
    };
    return "UNKNOWN_PORT_TYPE";
}


char* vhdl_create_components ( st_GDGELEMENTS_ALL *elmall, gboolean unused_elements ) {

    GString *str = g_string_new ( "" );

    int i;
    for ( i = 0; i < elmall->count_element; i++ ) {
        st_GDGELEMENT *elm = elmall->element[i];

        if ( ( elm->count_instance != 0 ) || unused_elements ) {
            g_string_append ( str, "\t--\n" );
            g_string_append_printf ( str, "\t-- count instances: %d\n", elm->count_instance );
            g_string_append ( str, "\t--\n" );
            g_string_append_printf ( str, "\tCOMPONENT %s_%s\n", elm->name, elm->description );
            g_string_append ( str, "\t\tPORT(\n" );

            int j;
            for ( j = 0; j < elm->count_port; j++ ) {
                st_GDGELEMENT_PORT *port = elm->port[j];
                g_string_append_printf ( str, "\t\t\t%s : %s STD_LOGIC%s\n", port->name, vhdl_get_port_type_txt ( port->type ), ( j == elm->count_port - 1 ) ? "" : ";" );
            };

            g_string_append ( str, "\t\t);\n" );
            g_string_append ( str, "\tEND COMPONENT;\n\n" );
        };
    };

    return g_string_free ( str, FALSE );
}


char* vhdl_create_signal_definitions ( st_GDGNETS_ALL *netall, const char *net_prefix ) {

    GString *str = g_string_new ( "" );
    GString *str_const = g_string_new ( "" );

    int i;
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];
        if ( !net->is_constant ) {
            g_string_append_printf ( str, "\tsignal %s%s : std_logic;\n", net_prefix, net->name );
        } else {
            g_string_append_printf ( str_const, "\tsignal %s : std_logic;\n", net->name );
        };
    };

    if ( str_const->len ) {
        g_string_append ( str, "\n\t-- constants\n" );
        g_string_append ( str, str_const->str );
    };

    g_string_free ( str_const, TRUE );
    return g_string_free ( str, FALSE );
}


char* vhdl_create_entity_port_connections ( st_GDGIOPORTS_ALL *ioportall, st_GDGNETS_ALL *netall, const char *net_prefix ) {
    GString *str_in = g_string_new ( "\n" );
    GString *str_out = g_string_new ( "\n" );
    GString *str_const = g_string_new ( "\n" );

    int i;
    for ( i = 0; i < ioportall->count_real; i++ ) {
        st_GDGIOPORT *ioport = ioportall->real[i];
        st_GDGNET *net = ioport->net;

        if ( net->count_output ) {
            g_string_append_printf ( str_out, "\t%s <= %s%s;\n", ioport->vhdl_name, net_prefix, net->name );
        } else {
            g_string_append_printf ( str_in, "\t%s%s <= %s;\n", net_prefix, net->name, ioport->vhdl_name );
        };
    };

    int count_const = 0;
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];
        if ( net->is_constant ) {
            count_const++;
            g_string_append_printf ( str_const, "\t%s <= '%d';\n", net->name, net->c_value );
        };
    };

    if ( count_const ) {
        g_string_append ( str_in, str_const->str );
    };
    g_string_free ( str_const, TRUE );

    g_string_append ( str_in, str_out->str );
    g_string_free ( str_out, TRUE );

    g_string_append ( str_in, "\n" );

    return g_string_free ( str_in, FALSE );
}


char* vhdl_create_component_instances ( st_GDGPARTS_ALL *partall, const char *part_prefix, const char *net_prefix ) {

    GString *str = g_string_new ( "" );


    int i;
    for ( i = 0; i < partall->count_part; i++ ) {
        st_GDGPART *part = partall->part[i];
        st_GDGELEMENT *elm = part->element;

        char *part_name = gdgpart_create_name ( part );
        g_string_append_printf ( str, "\t%s%s: %s_%s PORT MAP (\n", part_prefix, part_name, elm->name, elm->description );

        int j;
        for ( j = 0; j < elm->count_port; j++ ) {
            st_GDGNET *net = part->netpin[j];

            GString *str_signame = g_string_new ( "" );

            if ( net->is_constant ) {
                g_string_append ( str_signame, net->name );
            } else {
                g_string_append_printf ( str_signame, "%s%s", net_prefix, net->name );
            };

            g_string_append_printf ( str, "\t\t%s => %s%s\n", elm->port[j]->name, str_signame->str, ( j == elm->count_port - 1 ) ? "" : "," );
            g_string_free ( str_signame, TRUE );
        };

        g_string_append ( str, "\t);\n\n" );

        g_free ( part_name );
    };

    return g_string_free ( str, FALSE );
}
